import typing
import types
import logging
import json
import lxml
import lxml.etree
import abc
import re
import requests
import collections
import presalytics
import presalytics.lib.registry
import presalytics.lib.exceptions
import presalytics.lib.widgets.ooxml


logger = logging.getLogger(__name__)


class XmlTransformBase(abc.ABC):
    """
    Base class for writing Open Office Xml tranform functions to be implemented
    by `presalytics.lib.widgets.ooxml_editors.OoxmlEditorWidget`
    
    *For more information about Open Office Xml Schema that underlies 
    .pptx and .xlsx files, see http://officeopenxml.com/

    *For more information on how to lxml to write transforms, consult https://lxml.de/

    Parameters
    ----------
    function_params : dict
        A dictionary containing variables that will be used when the `transform_function`
        is executed
    """
    __xml_transform_kind__ = 'XmlTransform'

    def __init__(self, function_params: typing.Dict, *args, **kwargs):
        self.function_params = function_params

    @abc.abstractmethod
    def transform_function(self, lxml_element: lxml.etree.Element, params: typing.Dict) -> lxml.etree.Element:
        """
        Modifies Open Office Xml via transforming `lxml.etree.Element` using a params `dict` containing
        variables. Must be overridden in subclasses.
        """
        pass
    

    def execute(self, lxml_element: lxml.etree.Element):
        """
        Called by widget classes (e.g., `presalytics.lib.widgets.ooxml_editors.OoxmlEditorWidget`) to 
        perform the updates prescribed in the `transform_function`
        """
        return self.transform_function(lxml_element, self.function_params)


class ChangeShapeColor(XmlTransformBase):
    """
    Changes the color of a set of [Open Office Xml Shapes](http://officeopenxml.com/drwShape.php)
    
    Function Parameters Dictionary
    ----------
    hex_color : str
        The six-digit hexadecimal-format color string (e.g., ffa500 for orange).  See
        https://www.color-hex.com/ for an example color calculator

    object_name : str, optional
        The object tree name of the target shape.  If not supplied, all descendent shapes will
        have their color changed.      
    """
    __xml_transform_name__ = "ChangeShapeColor"

    @staticmethod
    def replace_color_on_target_shape(shape_xml: lxml.etree.Element, new_color: str) -> lxml.etree.Element:
        """
        Changes the color of an [Open Office Xml Shape](http://officeopenxml.com/drwShape.php)
        """
        new_fill = lxml.etree.Element("solidFill")
        lxml.etree.SubElement(new_fill, "srbgClr", {"val": new_color})
        fill_tags = [
            'noFill',
            'blipFill',
            'gradFill',
            'pattFill',
            'solidFill'
        ]
        props = shape_xml.find('.//{*}spPr')
        for child in props.getchildren():
            if any(re.search('{*}' + r, child.tag) for r in fill_tags):
                current_nsmap = child.nsmap
                current_index = props.index(child)
                current_prefix = child.prefix
                current_ns = current_nsmap.get(current_prefix, None)
                new_fill = lxml.etree.Element(lxml.etree.QName(current_ns, "solidFill"), nsmap=current_nsmap)
                lxml.etree.SubElement(new_fill, lxml.etree.QName(current_ns, "srgbClr"), {"val": new_color}, nsmap=current_nsmap)
                child.getparent().remove(child)
                props.insert(current_index, new_fill)
        return shape_xml

    def transform_function(self, lxml_element, params):
        """
        Changes the color of an [Open Office Xml Shape](http://officeopenxml.com/drwShape.php)

        Parameters
        -----------
        lxml_element : lxml.etree.Element
            An element containing as least one an `<sp>` element

        params : dict
            See the `Function Parameters Dictionary` for required entries
        """
        
        if re.match('{.*}sp', lxml_element.tag):
            shapes = [lxml_element]
        else:
            shapes = lxml_element.findall('.//{*}sp')
        object_name = params.get("object_name", None)
        color = params.get("hex_color")
        if object_name:
            for shape in shapes:
                is_target = False
                nvprops = shape.find('.//{*}nvSpPr')
                if len(nvprops) > 0:
                    name = nvprops.find('.//*[@name]').get("name")
                    if name == object_name:
                        is_target = True
                if is_target:
                    shape = ChangeShapeColor.replace_color_on_target_shape(shape, color)
        else:
            for shape in shapes:
                shape = ChangeShapeColor.replace_color_on_target_shape(shape, color)
        return lxml_element

class TextReplace(XmlTransformBase):
    """
    Replaces text in a template built into an Office Office Xml object that has 
    been uploaded to the Presalytics Ooxml Automation service.  Text to be should
    be in the format of a "template tag", which is a string enclosed in handlebars
    as such: '{{template_tag}}'

    Function Parameters Dictionary
    ----------

    A dictionary that maps template tags to the new strings that will replace the
    tags in the rendered widget.  The the dictionary keys should not be enclosed in handlebars.
    """
    __xml_transform_name__ = "TextReplace"

    class TextElementInfo(object):
        """
        Holds metadata for a list of `<t>` tags from an Office Open Xml document. 
        """
        def __init__(self, element: lxml.etree.Element, start_position: int):
            self.element = element
            self.text = "" if not element.text else element.text
            self.length = len(self.text)
            self.start_position = start_position
            self.end_position = self.start_position + self.length - 1

    class TextList(object):
        """
        Class for managing a list of `TextElementInfo` objects
        """
        _list: typing.List['TextReplace.TextElementInfo']

        def __init__(self):
            self._list = []

        def append(self, text_element_info):
            if isinstance(text_element_info, globals()['TextReplace'].TextElementInfo):
                self._list.append(text_element_info)
            else:
                raise presalytics.lib.exceptions.InvalidArgumentException(message="Argument must be an instance of class 'TextElementInfo'")
        
        def set_text(self, index, new_string):
            self._list[index].text = new_string
            self._list[index].element.text = new_string

        def get_text(self, index):
            return self._list[index].text

        def get_plaintext_string(self):
            ret = ""
            for item in self._list:
                ret += item.text
            return ret

        def get_position(self, match_string):
            return self.get_plaintext_string().find(match_string)

        def get_list_index_of_position(self, position: int):
            for item in self._list:
                if item.start_position <= position and item.end_position >= position:
                    return self.get_index(item)
            raise presalytics.lib.exceptions.InvalidArgumentException(message="Position {} out of range".format(position))
        
        def get_index(self, text_element_info: 'TextReplace.TextElementInfo'):
            for i in range(len(self._list)):
                if self._list[i] == text_element_info:
                    return i
            raise presalytics.lib.exceptions.InvalidArgumentException(message="Supplied argument on in self._list")
        
        def set_text_to_empty_string(self, index):
            self._list[index].text = ""
            self._list[index].element.text = ""

        def set_start_to_new_value(self, index, new_value):
            start_text, remainder = self._list[index].text.split("{{", 1)
            end = ""
            if "}}" in remainder:
                end = remainder.split("}}", 1)[1]
            new_text = "{0}{1}{2}".format(start_text, new_value, end)
            self._list[index].text = new_text
            self._list[index].element.text = new_text

        def truncate_end(self, index):
            end_text = self._list[index].text.split("}}", 1)[-1]
            self._list[index].text = end_text
            self._list[index].element.text = end_text

        def plaintext_string_list(self):
            return [x.text for x in self._list]

        def reset(self):
            for i in range(0, len(self._list)):
                start_position = 0 if i == 0 else self._list[i-1].end_position + 1
                self._list[i] = TextReplace.TextElementInfo(self._list[i].element, start_position)


            
    def replace_handlebars(self, info_list, params):
        """
        Method that finds template tags and replaces them 
        """
        for key, val in params.items():
            match_key = "{{" + key + "}}"
            match_start_position =  info_list.get_position(match_key)
            if match_start_position > -1:
                match_end_position = match_start_position + len(match_key) - 1
                match_start_index = info_list.get_list_index_of_position(match_start_position)
                match_end_index = info_list.get_list_index_of_position(match_end_position)
                if match_start_index < match_end_index:
                    info_list.truncate_end(match_end_index)
                for i in range(match_start_index + 1, match_end_index):
                    info_list.set_text_to_empty_string(i)
                info_list.set_start_to_new_value(match_start_index, val)
                info_list.reset()
                self.replace_handlebars(info_list, params)


    def transform_function(self, lxml_element, params):
        """
        Replaces template tags located {{inside_handlebars}} that match keys in the
        `params` dict with values from the `params` dict. 

        This method searches for match for plain text strings, so that if template
        tags are split across xml `<t>` elements, they are still identified and replaced. 

        Parameters
        -----------
        lxml_element : lxml.etree.Element
            An element containing as least one an `<t>` element

        params : dict
            A dictionary that maps template tags (no handlebars) to their respective
            replacement values.
        """
        text_list = lxml_element.findall('.//{*}t')
        info_list = TextReplace.TextList()
        position = 0
        for tag in text_list:
            info = self.TextElementInfo(tag, position)
            position = info.end_position + 1
            info_list.append(info)
        self.replace_handlebars(info_list, params)
        return lxml_element


class XmlTransformRegistry(presalytics.lib.registry.RegistryBase):
    """
    Registry for XmlTransform Classes.  Read by `presalytics.lib.widgets.ooxml_editors.MultiXmlTransform`
    and `presalytics.lib.widgets.ooxml_editors.OoxmlEditorWidget` so XmlTransformBase subclasses
    can be deserialized at run-time without a loaded instance in `locals()`.
    """
    def __init__(self):
        include_paths = presalytics.COMPONENTS.autodiscover_paths # todo: Change this line -does not exist at timport time.
        reserved_names = presalytics.COMPONENTS.reserved_names
        super(XmlTransformRegistry, self).__init__(autodiscover_paths=include_paths, reserved_names=reserved_names)

    def get_name(self, klass):
        return getattr(klass, "__xml_transform_name__", None)

    def get_type(self, klass):
        return getattr(klass, "__xml_transform_kind__", None)


XML_TRANSFORM_REGISTRY = None
"""
Static instance of `presalytics.lib.widgets.ooxml_editors.XmlTransformRegistry`. 

Should not be used directly by consuming classes. Only initialized if a consuming 
class or method calls the `presalytics.lib.widgets.ooxml_editors.get_transform_registry`.  This is done
for performance reasons since there's no need to build this registry at import-time if its never 
used in a workspace.
"""

def get_transform_registry():
    """
    Initalizes and retreives `presalytics.lib.widgets.ooxml_editors.XML_TRANSFORM_REGISTRY`
    """
    global XML_TRANSFORM_REGISTRY
    if not XML_TRANSFORM_REGISTRY:
        XML_TRANSFORM_REGISTRY = XmlTransformRegistry()
    return XML_TRANSFORM_REGISTRY



class MultiXmlTransform(XmlTransformBase):
    """
    This class allow users to run mutiple transforms on multiple targets in a single widget.  `MultiXmlTransform`
    wraps multiple subclasses of `presalytics.lib.widgets.ooxml_editors.XmlTransformBase`, creates instances of
    them, and feeds them into an `presalytics.lib.widgets.ooxml_editors.OoxmlEditorWidget` instance.  To do this,
    the `presalytics.lib.widgets.ooxml_editors.XmlTransformBase` subclasses must be loaded into the 
    `presalytics.lib.widgets.ooxml_editors.XML_TRANSFORM_REGISTRY` when called.

    The function parameters for the 

    Parameters
    ----------
    fail_quietly: bool, optional
        Defaults to false.  Indicates whether an exception should be raised when a subclass specified
        in the function parameters cannot not be found in the `presalytics.lib.widgets.ooxml_editors.XML_TRANSFORM_REGISTRY`
        instance. 
        

    Function Parameters Dictionary
    ----------
    transforms_list: list of dict
        A list of dictionaries, with each item in the list consiting of a dictionary of two entries.  
        The entries are as follows:
         
         * name: [str]
            The name of the the subclass of `presalytics.lib.widgets.ooxml_editors.XmlTransformBase` that
            will be that will be initialized
         
         * function_params: [dict]
            The `params` that will be the loaded into the instance's `transform_function` 
    """
    transform_instances: typing.List[XmlTransformBase]

    __xml_transform_name__ = "MultiXmlTransform"

    def __init__(self, transforms: typing.Dict[str, typing.List[typing.Dict[str, typing.Any]]], fail_quietly=True, **kwargs):
        super(MultiXmlTransform, self).__init__(transforms, **kwargs)
        transforms_list = transforms.get("transforms_list", None)
        self.fail_quietly = fail_quietly
        self.transform_instances = []
        self.transform_registry = get_transform_registry()
        for _transform in transforms_list: #type: ignore
            key = "XmlTransform." + _transform["name"]
            transform_class = self.transform_registry.get(key)
            if not transform_class:
                message = "Could not find XmlTransform with name '{}'".format(_transform["name"])
                if self.fail_quietly:
                    logging.info(message)
                else:
                    raise self.transform_registry.raise_error(message) #noqa
            else:
                inst = transform_class(_transform["function_params"])
                self.transform_instances.append(inst)

    def transform_function(self, lxml_element, params):
        for inst in self.transform_instances:
            lxml_element = inst.execute(lxml_element)
        return lxml_element    


class OoxmlEditorWidget(presalytics.lib.widgets.ooxml.OoxmlWidgetBase):
    """
    Edits a `widget` from a Presentation or Spreadsheet document and renders 
    the edited widget.

    This class interacts with the Presalytics API to extract SVG objects from
    Presentation and spreadsheet documents, from Presaltytics Ooxml Automation
    service objects that have already been loaded into the API.  This class requires
    that users supply `transform_function` by subclassing 
    `presalytics.lib.widgets.ooxml_editors.XmlTransformBase`, and an optional set of 
    parameters to act as variables in the transform function.

    Parameters
    ----------
    filename : str
        The local filepath a presentation or spreadsheet file containing
        the object to be rendered

    name : str
        The widget name.  If not provided, attribute will be set as the `object_name` 
        or `filename`

    story_id : str
        The the id of the story in the Presalytics API Story service.  If not provided, 
        a new story will be created.  Do not supply if this object has not yet been created. 
    
    object_ooxml_id : str
        The identifier of the Ooxml Automation service object bound the Story. Do not supply if this 
        object has not yet been created.

    endpoint_map : presalytics.lib.widgets.ooxml.OoxmlEndpointMap
        Reference to the Presalytics API Ooxml Automation service endpoint and object type
        that for the object of interest

    tranform_class : subclass of presalytics.lib.widgets.ooxml_editors.XmlTransformBase
        A class containing a `transform_function` method that transforms Open Office Xml via
        an `lxml.etree.Element` instance

    transform_params : dict, optional
        A dictionary of parameters that will be passed to the `transform_class`'s `transform_function`
        as variables to modify the underlying OpenOfficeXml 
    
    """
    transform: XmlTransformBase

    __component_kind__ = 'ooxml-xml-editor'

    def __init__(self,
                 name: str,
                 story_id,
                 object_ooxml_id,
                 endpoint_map,
                 transform_class,
                 transform_params={},
                 **kwargs):
        super(OoxmlEditorWidget, self).__init__(name, story_id=story_id, object_ooxml_id=object_ooxml_id, endpoint_map=endpoint_map, **kwargs)
        self.name = name
        self.transform = transform_class(transform_params)
        self.update()
        self.svg_html = self.create_container(**self.client_info)
        self.outline_widget = self.serialize()

    def update_xml(self, xml_str) -> str:
        """
        Runs the `transform_function` Open Office Xml data downloaded via the Presalytics API
        """
        xml = lxml.etree.fromstring(xml_str)
        new_xml = self.transform.execute(xml)
        new_xml_str = lxml.etree.tostring(new_xml)
        return new_xml_str

    def update(self):
        """
        Update the widget, include changes to the Xml
        """
        client = self.get_client()
        headers = client.get_auth_header()
        headers.update(client.get_request_id_header())
        if self.transform:
            xml_url = self.endpoint_map.get_xml_url(self.object_ooxml_id)
            xml_response = requests.get(xml_url + "?updated=false", headers=headers)
            if xml_response.status_code != 200:
                raise presalytics.lib.exceptions.ApiError(message=xml_response.text)
            dto = xml_response.json()
            new_xml = self.update_xml(dto["openOfficeXml"])
            dto["openOfficeXml"] = new_xml.decode('utf-8')
            xml_update_response = requests.put(xml_url, json=dto, headers=headers)
            if xml_response.status_code != 200:
                raise presalytics.lib.exceptions.ApiError(message=xml_update_response.content)

    @classmethod
    def deserialize(cls, component, **kwargs):
        endpoint_map = presalytics.lib.widgets.ooxml.OoxmlEndpointMap(component.data["endpoint_id"])
        class_key = "XmlTransform." + component.data.get("transform_class", "")
        transform_class = get_transform_registry().get(class_key)
        transform_params = component.data.get("transform_params", {})
        return cls(component.name,
                   component.data["story_id"],
                   component.data["object_ooxml_id"],
                   endpoint_map,
                   transform_class=transform_class,
                   transform_params=transform_params,
                   **kwargs)

    def serialize(self, **kwargs):
        data = {
            "story_id": self.story_id,
            "object_ooxml_id": self.object_ooxml_id,
            "endpoint_id": self.endpoint_map.endpoint_id,
            "transform_class": self.transform.__xml_transform_name__,
            "transform_params": self.transform.function_params
        }
        widget = presalytics.story.outline.Widget(
            name=self.name,
            data=data,
            kind=self.__component_kind__
        )
        return widget



import typing
import logging
import urllib.parse
import requests
import presalytics.lib.tools.ooxml_tools
import presalytics.lib.exceptions
if typing.TYPE_CHECKING:
    from presalytics.client.api import Client
    from io import BytesIO


logger = logging.getLogger(__name__)


def story_post_file_bytes(client: 'Client', 
                          binary_obj: 'BytesIO', 
                          filename: str,
                          mime_type: str = None):
    """
    Create a Presalytics API Story object from a file-like `io.BytesIO` object.  Helpful for server-side 
    interaction with the Presalytics Story API

    Parameters
    ----------
    client : presalytics.client.api.Client
        A client object for making api calls
    
    binary_obj : io.BytesIO
        A file-like object for storing file-data in memory.  Often found in multipart messages
        uploaded from browsers.
    
    filename : str
        The filename of the object to be uploaded

    mimetype : str, optional
        If known, please add the mimetype of the file.  Otherwise, this method will execute an 
        additional API call ascertain the file's mimetype

    Returns
    ----------
    A `presalytics.client.presalytics_story.models.story.Story` containing information about the Story object in the Presalytics API
    """
    if not mime_type:
        mime_type = presalytics.lib.tools.ooxml_tools.get_mime_type_from_filename(client, filename)
    _file = {'file': (filename, binary_obj, mime_type,)}
    headers = client.get_auth_header()
    headers.update(client.get_request_id_header())
    headers.update({
        'User-Agent': client.story.api_client.user_agent,
        'Accept': 'application/json'
    })
    endpoint = urllib.parse.urljoin(client.story.api_client.configuration.host, 'story/file')
    try:
        resp = requests.post(endpoint, headers=headers, files=_file)
    except Exception as ex:
        message = "An error occured in the presalytics API client"
        if locals().get("resp", None):
            code = resp.status_code
        else:
            code = 500
        raise presalytics.lib.exceptions.ApiError(message=message, status_code=code)
            
    data = resp.json()
    if resp.status_code > 299:
        logger.error(data['detail'])
        raise presalytics.lib.exceptions.ApiError(message=data["detail"], status_code=resp.status_code)
    else:
        try:
            story = client.story.api_client._ApiClient__deserialize(data, 'Story')
            return story
        except Exception as ex:
            logger.error("Story object could not be deserialized.")
            logger.exception(ex)
        return data



"""
Command Line Interface for Presalytics Python Library

Run "presalytics -h" for details
"""
import os
import logging
import argparse
import presalytics
import urllib.parse
import webbrowser
import yaml
import json
import ast
import datetime
import presalytics.lib.constants
import presalytics.lib.tools.story_tools
import presalytics.lib.tools.ooxml_tools
import presalytics.lib.tools.workflows


logger = logging.getLogger(__name__)

description = """
Presalytics Python Library
--------------------------
Version: {version}

Create live presentations, dashboards, and persistent analytics,
and interact with the Presalytics API.

Please review the push, pull, and create subcommands for more options.

For more information about the Presalytics API, please visit 
<https://presalytics.io> or send your questions to inquires@presalytics.io.

Command Line Instructions
-------------------------
""".format(version=presalytics.__version__)

parser = argparse.ArgumentParser(
    description=description,
    epilog="Further documentation is available at <https://presalytics.io/docs/>.",
    formatter_class=argparse.RawTextHelpFormatter,
    prog='presalytics'
)
file_help = "The filename of the Story Outline. Defaults to 'story.yaml'"
parser.add_argument('-f', '--file', default='story.yaml', action='store', help=file_help)
parser.add_argument('--version', action='version', version=presalytics.__version__)
parser.add_argument('--view', default=False, action='store_true', help='View a Story in the browser')
parser.add_argument('--manage', default=False, action='store_true', help="""Go to the Story's management page""")
parser.add_argument('--show-story', default=False, action='store_true', help="Print Story metadata to the console")
parser.add_argument('--cron', default=False, action='store_true', help="Print information on how automate updates to your story to the console.")
verbosity = parser.add_mutually_exclusive_group(required=False)
verbosity.add_argument('-v', '--verbose', default=False, action='store_true', help="Increases the detail in log output")
verbosity.add_argument('-q', '--quiet', default=False, action='store_true', help="Decrease the detail in log output. Supresseses non-critical errors.")


yaml_help = "Writes file updates to YAML format.  Yaml is the default."
json_help = "Writes file updates to JSON format"
overwrite_help = "Forces (o)verwrite of file with returned Story Outline (if exists)"
username_help = "Overrides the username in presalytics.CONFIG (if present)"
password_help = "The user's Presalytics API password"
subparsers = parser.add_subparsers(title='Story API Commands', prog='presalytics', dest='story_api')

push_description = """
Pushes a Story Outline revision to the Presalytics API
Story service, and writes the updated story outline to a local file (--file)
"""
push = subparsers.add_parser('push', description=push_description, help='Push a Story Outline revision')
push.add_argument('--update', default=False, action='store_true', help="Update the Story Outline from Local Scripts before pushing")
push.add_argument('-o', '--overwrite', default=False, action='store_true', help=overwrite_help)
push.add_argument('-u', '--username', default=None, action='store', help=username_help)
push.add_argument('-p', '--password', default=None, action='store', help=password_help)
push.add_argument('-m', '--message', default=None, action='store', help="Revision message (added to outline).  Automatically populates by default.")
push_options = push.add_mutually_exclusive_group(required=False)
push_options.add_argument('-y', '--yaml', default=False, action='store_true', help=yaml_help)
push_options.add_argument('-j', '--json', default=False, action='store_true', help=json_help)

pull_description = """
Retrives a Story Outline from the the Presalytics API Story Service,
and writes the retrieved story outline to a local file (--file)
"""

id_help = """
The Preslytics API Story Service Id for the story (type: UUID-v4)
If not supplied, the tool searches the file from the --file option 
for a 'storyId' attribute 
"""

pull = subparsers.add_parser('pull', description=pull_description, help='Pull a Story Outline revision')
pull.add_argument('--id', default=None, action='store', help=id_help)
pull.add_argument('-o', '--overwrite', default=False, action='store_true', help=overwrite_help)
pull.add_argument('-u', '--username', default=None, action='store', help=username_help)
pull.add_argument('-p', '--password', default=None, action='store', help=password_help)
pull_options = pull.add_mutually_exclusive_group(required=False)
pull_options.add_argument('-y', '--yaml', default=False, action='store_true', help=yaml_help)
pull_options.add_argument('-j', '--json', default=False, action='store_true', help=json_help)


create_description = """

The widget or page instance in the [name] arguemnt must be avialable in 
`presalytics.COMPONENTS` at run-time.  Widget and page instances are loaded 
into `presalytics.COMPONENTS` from the current working directory and other 
configured folders at import of the presalytics module.
"""

create_epilog = """
See https://presalytics.io/docs/configuration/ for more information.
"""

name_help = """
The name of the widget or page instance to create a Story Outline from.  If using options
`ooxml_file`, the name is the path the file that you would like to create from.
"""
create = subparsers.add_parser('create', description=create_description, help='Create a Story Outline', epilog=create_epilog)
create.add_argument('name', action='store', help=name_help)
create_options = create.add_mutually_exclusive_group(required=True)
create_options.add_argument('--widget', default=False, action='store_true', help="Create a Story Outline from a widget instance")
create_options.add_argument('--page', default=False, action='store_true', help="Create a Story Outline from a page instance")
create_options.add_argument('--ooxml_file', default=False, action='store_true', help="Create a Story Outline from an Ooxml Document")
create.add_argument('-o', '--overwrite', default=False, action='store_true', help=overwrite_help)
create.add_argument('-u', '--username', default=None, action='store', help=username_help)
create.add_argument('-p', '--password', default=None, action='store', help=password_help)
create.add_argument('-s', '--source', default=None, action='store', help="The module containing the instance.  Needed only if instance not auto-loaded into presalytics.CONFIG")
create_output_options = create.add_mutually_exclusive_group(required=False)
create_output_options.add_argument('-y', '--yaml', default=False, action='store_true', help=yaml_help)
create_output_options.add_argument('-j', '--json', default=False, action='store_true', help=json_help)

update_description = """
Update the Story outline from instances contained in scripts in the active workspace 
"""

update = subparsers.add_parser('update', description=update_description, help='Update a Story Outline From Local Scripts')
update.add_argument('-m', '--message', default=None, action='store', help="Revision message (added to outline).  Automatically populates by default.")
update_options = update.add_mutually_exclusive_group(required=False)
update_options.add_argument('-y', '--yaml', default=False, action='store_true', help=yaml_help)
update_options.add_argument('-j', '--json', default=False, action='store_true', help=json_help)


modify_description = """
Modify a Story's outline

You can either 'add' or 'remove' a widget to or from a page in story outline.

For more complex operations, you can apply a JSON 'patch' to to the story outline per 
[RFC 6902](https://tools.ietf.org/html/rfc6902).  You can find good exmaples at 
www.jsonpatch.com
""" 
modify = subparsers.add_parser('modify', description=modify_description, help='Modify a Story Outline')
modify.add_argument('action', choices=['add', 'remove', 'patch'], action='store', help="You can either add or remove a widget (quick & easy), or apply a json patch (more complex)")
modify.add_argument('-n', '--name', default=None, action='store', help="The name of the widget you would like to add or remove")
modify.add_argument('--position', default=None, action='store', type=int, help="The position in the widget list to place the widget")
modify.add_argument('--page_number', default=None, action='store', type=int, help="The page number to add or remove the widget to/from" )
modify.add_argument('--patch', default=None, action='store', help="The json patch (per RFC 6902) you want to apply to the Story Outline.")
modify_output_options = modify.add_mutually_exclusive_group(required=False)
modify_output_options.add_argument('-y', '--yaml', default=False, action='store_true', help=yaml_help)
modify_output_options.add_argument('-j', '--json', default=False, action='store_true', help=json_help)

share_description = """
Control permissions for owners, editors, promoters, and viewer or your stories
"""

share = subparsers.add_parser('share', description=share_description, help='Share Stories')
share.add_argument('--user-ids', action='store', help='A comma-separated list of the Presalytics API user IDs of users you what to share this story with.')
share.add_argument('--emails', action='store', help='A comma-separated list of the email addresses of users you what to share this story with.')
share.add_argument('--collaborator-type', choices=['owner', 'viewer', 'promoter', 'editor'], action='store', help='The permission type to grant to the new user. Defaults to viewer.')
share.add_argument('-u', '--username', default=None, action='store', help=username_help)
share.add_argument('-p', '--password', default=None, action='store', help=password_help)


account_description = """
Utilities for managing and you account and your stories.  Use the [--delete]
option to delete stories
"""

account = subparsers.add_parser('account', description=account_description, help='Account & Story Management Utlities')
account_action_group = account.add_mutually_exclusive_group(required=True)
account_action_group.add_argument('--delete', default=False, action='store_true', help='Delete Stories')
delete_group = account.add_mutually_exclusive_group(required=True)
delete_group.add_argument('--all', default=False, action='store_true', help='WARNING!!! Delete all stories on account.')
delete_group.add_argument('--id', default=None, help='The Presalytics API Story service Id of the story you want to delete.')
account.add_argument('-u', '--username', default=None, action='store', help=username_help)
account.add_argument('-p', '--password', default=None, action='store', help=password_help)

ooxml_description = """
Utilities for managing references to the Presaltyics API Ooxml Automation Service in your stracts
"""
ooxml = subparsers.add_parser('ooxml', description=account_description, help='Modify Stories using Ooxml Documents')
ooxml.add_argument('ooxml_filepath', action='store', default=None, help="The relative or absolute file path to the ooxml-file")
ooxml.add_argument('action', choices=['add', 'replace'], default=None, action='store', help="Whether to add the ooxml to a story or replace an existing one.")
ooxml.add_argument('--story-id', action='store', default=None, help="The Presalytics API Story service Id of the story you want associate this file with.  Defaults to the story at the [--file] option.")
ooxml.add_argument('--replace-id', action='store', default=None, help="The Ooxml Automation service if id for the associated document that you want to replace")
ooxml.add_argument('-u', '--username', default=None, action='store', help=username_help)
ooxml.add_argument('-p', '--password', default=None, action='store', help=password_help)

config_description = """
Create and manage presalytics `config.py` files
"""

config = subparsers.add_parser('config', description=config_description, help='Create and manage presalytics `config.py` files')
config.add_argument('username', action='store', help=username_help)
config.add_argument('-p', '--password', default=None, action='store', help=password_help)
config.add_argument('-s', "--set", metavar="KEY=VALUE", default=None, nargs='+', help="Pass config values to to `config.py` with KEY=VALUE stucture (e.g., '-s USE_LOGGER=False'")
config.add_argument('-o', '--overwrite', default=False, action='store_true', help=overwrite_help)

def parse_var(s):
    """
    Parse a key, value pair, separated by '='
    That's the reverse of ShellArgs.

    On the command line (argparse) a declaration will typically look like:
        foo=hello
    or
        foo="hello world"
    """
    items = s.split('=')
    key = items[0].strip() # we remove blanks around keys, as is logical
    if len(items) > 1:
        # rejoin the rest:
        value = '='.join(items[1:])
        if value == 'True' or value == 'true':
            value = True
        if value == 'False'or value == 'false':
            value = False
    return (key, value)


def parse_vars(items):
    """
    Parse a series of key-value pairs and return a dictionary
    """
    d = {}

    if items:
        for item in items:
            key, value = parse_var(item)
            d[key] = value
    return d

def _load_file(filename):
    if filename.endswith('yaml') or filename.endswith('yml'):
        outline = presalytics.StoryOutline.import_yaml(filename)
    if filename.endswith('json'):
        with open(filename, 'r') as f:
            outline_string = f.read()
        outline = presalytics.StoryOutline.load(outline_string)
    return outline


def _make_url(story_id, url_type):
    route = "/story/{0}/{1}/".format(url_type, story_id)
    try:
        host = presalytics.CONFIG["HOSTS"]["SITE"]
    except (KeyError, AttributeError):
        host = presalytics.lib.constants.SITE_HOST
    return urllib.parse.urljoin(host, route)

def _open_page(story_id, url_type):
    url = _make_url(story_id, url_type)
    webbrowser.open_new_tab(url)

def _write(outline, filename, json=False):
    if json:
        with open(filename, 'w') as f:
            f.write(outline.dump())
    else:
        outline.export_yaml(filename)
    
def _dump(outline, filename, overwrite=False, json=False):
    if os.path.exists(filename):
        if not overwrite:
            logger.error("A file already exists at {}.  Use option [--overwrite] to overwrite the existing file".format(filename))
        else:
            os.remove(filename)
            _write(outline, filename, json)
    else:
        _write(outline, filename, json)


def main():
    """ Command-line entry point 
    
    Run the following from the command line for more information:

        python3 -m presalytics --help
    
    or inside a python virtual environment:

        presalytics -h
    
    """
    try:
        args = parser.parse_args()
        filename = args.file
        file_extension = filename.split(".")[-1]
        if file_extension == "yaml" or file_extension == "yml":
            original_file_is_yaml = True
        else:
            original_file_is_yaml - False
        lgs = [logging.getLogger(n) for n in logging.root.manager.loggerDict]
        if args.verbose or args.quiet:
            for lg in lgs:
                lg.setLevel(logging.DEBUG)
        elif args.quiet:
            for lg in lgs:
                lg.setLevel(logging.CRITICAL)
        else:
            for lg in lgs:
                lg.setLevel(logging.ERROR)
        write = False
        pull = False
        push = False
        account = False
        share = False
        outline = None
        ooxml = False
        config = False
        modify = False
        update = False
        message = getattr(args, "message", None)
        if args.story_api == "create":
            # create outline from page/widget
            if args.ooxml_file:
                ooxml_file = args.name
                if not os.path.exists(ooxml_file):
                    cwd = os.getcwd()
                    test_path = os.path.join(cwd, ooxml_file)
                    if os.path.exists(test_path):
                        ooxml_file = test_path
                    else:
                        logger.error("Could not find a path to file: {0}".format(ooxml_file))
                        return
                story = presalytics.lib.tools.ooxml_tools.create_story_from_ooxml_file(ooxml_file)
                outline = presalytics.story.outline.StoryOutline.load(story.outline)
            else:
                outline = presalytics.lib.tools.workflows.create_from_instance(args.name, page=args.page, widget=args.widget, filename=args.source)
            # dump to file
            push = True
            pull = True
            write = True
            message = outline.info.revision_notes
        elif args.story_api == "push":
            push = True
            pull = False
            write = False
            if args.update:
                update = True
        elif args.story_api == "pull":
            push = False
            pull = True
            write = True
        elif args.story_api == "account":
            account = True
        elif args.story_api == "share":
            share = True
        elif args.story_api == "ooxml":
            ooxml = True
            pull = True
        elif args.story_api == "config":
            config = True
        elif args.story_api == "modify":
            modify = True
        elif args.story_api == "update":
            update = True
        else:
            push = False
            pull = False
            write = False
            #load story outline from file
        if config:
            set_dict = {} if not args.set else parse_vars(args.set)
            presalytics.lib.tools.workflows.create_config_file(args.username, 
                                                               password=args.password, 
                                                               set_dict=set_dict, 
                                                               overwrite=args.overwrite)
            logger.info("File 'config.py creating in folder " + os.getcwd())
            return
        try:
            if not outline:
                if not os.path.exists(filename):
                    current_dir = os.getcwd()
                    abs_filename = os.path.join(current_dir, filename)
                    if os.path.exists(abs_filename):
                        filename = abs_filename
                    else:
                        logger.error("Could not find file: {}".format(filename))
                        return
                outline = _load_file(filename)
        except Exception:
            logger.error("Error handling file: {}".format(filename))
            return
        if modify:
            if args.action == "add" or args.action == "remove":
                if not args.name:
                    logger.error("Modifying an outline using the 'add' or 'remove' actions requires a [--name] argument")
                    return
                if args.action == "add":
                    outline = presalytics.lib.tools.workflows.add_widget_instance(args.name, outline, position=args.position, page_number=args.page_number, filename=filename)
                elif args.action == "remove":
                    outline = presalytics.lib.tools.workflows.remove_widget_by_name(args.name, outline, page_number=args.page_number, filename=filename)
            elif args.action == "patch":
                if not args.patch:
                    logger.error("Modifying an outline using the 'patch' action requires a [--patch] argument")
                    return
                try:
                    try:
                        patch = json.loads(args.patch)
                    except json.JSONDecodeError:
                        patch = ast.literal_eval(args.patch)
                except Exception as ex:
                    logger.error("A patch could not be created from [--patch]: {}".format(args.patch))
                    return
                outline = presalytics.lib.tools.workflows.apply_json_patch(outline, patch)
            _dump(outline, filename, True, args.json)
        if update:
            outline = presalytics.lib.tools.workflows.update_outline(outline, filename=filename, message=args.message)      
            _dump(outline, filename, True, args.json)  
        if push:
            if not message:
                pretty_time = datetime.datetime.now().strftime("%d-%m-%Y at %H:%M")
                message = "Pushed from command line interface on " + pretty_time
            outline.info.revision_notes = message
            outline = presalytics.lib.tools.workflows.push_outline(outline, username=args.username, password=args.password)
        try:
            story_id = outline.story_id
        except:
            logger.error("A story outline could not be found or created.  Please use the [--file] option to designate a target outline.")
            return
        if ooxml:
            if args.story_id:
                story_id = args.story_id
            if not args.replace_id and args.action == "replace":
                logger.error("the [--replace-id] option is required when the 'action' argument is [replace]")
                return
            ooxml_file = args.ooxml_filepath
            if not os.path.exists(ooxml_file):
                cwd = os.getcwd()
                test_path = os.path.join(cwd, ooxml_file)
                if os.path.exists(test_path):
                    ooxml_file = test_path
                else:
                    logger.error("Could not find a path to file: {0}".format(ooxml_file))
                    return
            if args.action == "add" or args.action == "replace": 
                presalytics.lib.tools.ooxml_tools.add_ooxml_document_to_story(story_id, ooxml_file, replace_id=args.replace_id, username=args.username, password=args.password)
        if pull:
            if story_id == "empty":
                logger.error("A story outline needs a Story Id to be pulled from the Presalytics API. Please run 'presalytics push'")
                return
            else:
                if getattr(args, "id", None):
                    _id = args.id
                else:
                    _id = outline.story_id
                outline = presalytics.lib.tools.workflows.pull_outline(_id, username=args.username, password=args.password)
       
        if write:
            _dump(outline, filename, args.overwrite, args.json)
        if account:
            if args.delete:
                if args.all:
                    presalytics.lib.tools.workflows.delete_all_stories(username=args.username, password=args.password)
                else:
                    presalytics.lib.tools.workflows.delete_by_id(args.id, username=args.username, password=args.password)
        if share:
            presalytics.lib.tools.workflows.share_story(story_id, 
                                                           emails=args.emails, 
                                                           user_ids=args.user_ids, 
                                                           username=args.username,
                                                           password=args.password,
                                                           collaborator_type=args.collaborator_type)
        if story_id != 'empty':
            try:
                if args.view:
                    _open_page(story_id, "view")
                if args.manage:
                    _open_page(story_id, "manage")
                if args.show_story:
                    story = presalytics.lib.tools.workflows.get_story(story_id)
        
            except webbrowser.Error:
                logger.error("This environment does not have a webrowser loaded for use with python.")
                return
        else:
            logger.error("This outline does not yet have a story_id.  Please run 'presalytics push'.")
            return
        logger.info("\n\nStory Outline\n-------------\n\n{0}".format(yaml.dump(outline.to_dict())))
        if args.cron:
            presalytics.lib.tools.workflows.create_cron_target()
    except Exception as ex:
        if isinstance(ex, presalytics.lib.exceptions.PresalyticsBaseException):
            logger.error(ex.message) # noqa
        else:
            logger.exception(ex)
    finally:
        return


if __name__ == "__main__":
    main()


# coding: utf-8

"""
    OOXML Automation

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


from setuptools import setup, find_packages  # noqa: H301

NAME = "presalytics_ooxml_automation"
VERSION = "0.2.92"
# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = ["urllib3 >= 1.15", "six >= 1.10", "certifi", "python-dateutil"]


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name=NAME,
    version=VERSION,
    description="OOXML Automation",
    author_email="kevin@presalytics.io",
    url="https://github.com/presalytics/ooxml-automation-python-client",
    keywords=["OpenAPI", "OpenAPI-Generator", "OOXML Automation"],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description_content_type='text/markdown',
    long_description=long_description
)
